/*
 * Copyright (c) 2000, 2006
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
#ifndef lint
static const char rcsid[] =
    "@(#) $Id: dhcpdconf.c 133 2008-10-25 21:38:58Z leres $ (LBL)";
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>

#include <netinet/in.h>

#include <arpa/inet.h>

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <memory.h>
#include <netdb.h>
#include <stdio.h>
#include <string.h>
#include <syslog.h>
#include <unistd.h>

#include "gnuc.h"
#ifdef HAVE_OS_PROTO_H
#include "os-proto.h"
#endif

#include "dhcpdconf.h"
#include "util.h"

/* Forwards */
static const char *addhash(u_int32_t);
static const char *parsedhcpdconf(const char *, size_t);
static const char *checkdhcpdconf(int);
static void inithash(void);

#define HASHSIZE 4096
#define HASH(a) (&addr_table[(a) & (HASHSIZE - 1)])

static u_int32_t addr_table[HASHSIZE];

static char *dhcpdconf;

const char *
setdhcpdconf(char *file, int mustexist)
{

	dhcpdconf = file;
	return (checkdhcpdconf(mustexist));
}


/*
 * Determine if this is a "good client" or not.
 * Good clients have a dhcpdconf entry.
 */
static const char *
checkdhcpdconf(int mustexist)
{
	const char *bytes, *msg;
	int fd;
	struct stat st;
	size_t size;
	static char errmsg[128];
	static time_t mtime = 0, ctime = 0;

	if (dhcpdconf == NULL)
		return ("no dhcpdconf set");

	/* Stat the dhcpdconf */
	errno = 0;
	if (stat(dhcpdconf, &st) < 0) {
		if (!mustexist && errno == ENOENT)
			return ("");
		(void)sprintf(errmsg, "stat on \"%s\": %s",
		    dhcpdconf, strerror(errno));
		return (errmsg);
	}

	/* Nothing to do if nothing has changed */
	if (mtime == st.st_mtime && ctime == st.st_ctime)
		return (NULL);

	lg(LOG_INFO, "%sreading \"%s\"", mtime != 0L ? "re-" : "", dhcpdconf);

	/* Open dhcpdconf file */
	if ((fd = open(dhcpdconf, O_RDONLY)) < 0) {
		(void)sprintf(errmsg, "error opening \"%s\": %s",
		    dhcpdconf, strerror(errno));
		return (errmsg);
	}

	/*
	 * Record file modification time.
	 */
	if (fstat(fd, &st) < 0) {
		(void)sprintf(errmsg, "fstat: %s", strerror(errno));
		close(fd);
		return (errmsg);
	}
	mtime = st.st_mtime;
	ctime = st.st_ctime;
	size = st.st_size;

	bytes = mmap(NULL, size, PROT_READ, 0, fd, (off_t)0);
	if (bytes == MAP_FAILED) {
		(void)sprintf(errmsg, "mmap: %s", strerror(errno));
		close(fd);
		return (errmsg);
	}

	msg = parsedhcpdconf(bytes, size);
	close(fd);
	return (msg);
}

static const char fixedaddress[] = "fixed-address";

static const char *
parsedhcpdconf(const char *bytes, size_t size)
{
	int i, n;
	const char *cp, *cp2, *ep;
	char **ap;
	u_int32_t a;
	char addr[128];
	struct hostent *hp;

#define EATWHITESPACE \
	while (cp < ep) { \
		if (*cp == '\n') \
			++n; \
		else if (!isblank(*cp)) \
			break; \
		++cp; \
	}

	inithash();
	n = 0;
	cp = bytes;
	ep = bytes + size;
	while (cp < ep) {
		/* Eat comments */
		EATWHITESPACE
		if (*cp == '#') {
			do {
				++cp;
			} while (cp < ep && *cp != '\n');
			continue;
		}

		i = sizeof(fixedaddress) - 1;
		if (cp + i < ep && strncmp(cp, fixedaddress, i) == 0) {
			cp += i;
			cp2 = cp;
			EATWHITESPACE
			if (cp2 == cp) {
				lg(LOG_ERR,
				    "parsedhcpdconf: %s.%d: missing whitespace",
				    dhcpdconf, n);
				continue;
			}
			cp2 = cp;
			while (cp < ep &&
			    !isblank(*cp) && *cp != ';' && *cp != '\n')
				++cp;
			i = cp - cp2;
			if (i > sizeof(addr) - 1)
				i = sizeof(addr) - 1;
			strncpy(addr, cp2, i);
			addr[i] = '\0';

			a = inet_addr(addr);
			if ((int32_t)a == -1 || a == 0) {
				hp = gethostbyname(addr);
				if (hp == NULL) {
					lg(LOG_ERR,
					    "parsedhcpdconf: %s.%d:"
					    " bad host \"%s\"",
					    dhcpdconf, n, addr);
					continue;
				}
				for (ap = hp->h_addr_list; *ap != NULL; ++ap) {
					memmove(&a, *ap, sizeof(a));
					cp2 = addhash(a);
					if (cp2 != NULL)
						return (cp2);
				}
			} else {
				/* Got a live one */
				cp2 = addhash(a);
				if (cp2 != NULL)
					return (cp2);
			}
			continue;
		}
		if (*cp == '\n')
			++n;
		++cp;
	}
	return (NULL);
}

/* Returns true if the specified address is a client in the dhcpdconf file */
int
dhcpclient(u_int32_t a)
{
	const char *msg;
	int i;
	u_int32_t *p;

	msg = checkdhcpdconf(1);
	if (msg != NULL) {
		/* Empty string means no dhcpdconf but none required */
		if (*msg == '\0')
			return (0);
		lg(LOG_ERR, "goodclient: checkdhcpdconf: %s", msg);
	}

	p = HASH(a);
	p = &addr_table[a & (HASHSIZE - 1)];
	i = HASHSIZE;
	while (*p != 0 && i-- > 0) {
		if (*p == a)
			return (1);
		++p;
		if (p > &addr_table[HASHSIZE - 1])
			p = addr_table;
	}
	return (0);
}

static void
inithash(void)
{

	memset(addr_table, 0, sizeof(addr_table));
}

static const char *
addhash(u_int32_t a)
{
	int i;
	u_int32_t *p;

	p = HASH(a);
	i = HASHSIZE;
	while (*p != 0 && i-- > 0) {
		++p;
		if (p > &addr_table[HASHSIZE - 1])
			p = addr_table;
	}
	/* Check if the table is full */
	if (*p != 0)
		return ("parsedhcpdconf: hash table full");

	*p = a;
	return (NULL);
}
